import json
from abc import ABC, abstractmethod
from functools import wraps
from typing import Callable, Any

import jsonschema
from wai.common.decorator import ensure_error_type
from wai.common.meta import does_not_raise, instanceoptionalmethod

from ..error import JSONValidationError
from ..raw import RawJSONElement, deep_copy
from ..schema import JSONSchema


class JSONValidator(ABC):
    """
    Interface for classes which can validate raw JSON.
    """
    @instanceoptionalmethod
    def with_validation(self, function: Callable[[Any], RawJSONElement]) -> Callable[[Any], RawJSONElement]:
        """
        Decorator which adds validation to a function returning raw JSON.

        :param function:    The function to decorate.
        :return:            The decorated function.
        """
        @wraps(function)
        def validated(*args, **kwargs) -> RawJSONElement:
            # Call the function
            raw_json: RawJSONElement = function(*args, **kwargs)

            # Perform validation
            self.validate_raw_json(raw_json)

            return raw_json

        return validated

    @instanceoptionalmethod
    @ensure_error_type(JSONValidationError, "Error validating raw JSON element {raw_json}: {0}")
    def validate_raw_json(self, raw_json: RawJSONElement):
        """
        Validates the raw JSON using this object's validation methods
        (schema, special).

        :param self:        The validator instance/class.
        :param raw_json:    The raw JSON to validate.
        """
        # Get the jsonschema validator
        validator = self.get_validator()

        # Perform schema validation
        validator.validate(raw_json)

        # Perform special validation
        self.perform_special_json_validation(raw_json)

    @instanceoptionalmethod
    def is_valid_raw_json(self, raw_json: RawJSONElement) -> bool:
        """
        Checks if the raw JSON is valid.

        :param self:        The validator instance/class.
        :param raw_json:    The raw JSON to check.
        :return:            True if the JSON is valid,
                            False if not.
        """
        return does_not_raise(self.validate_raw_json, raw_json)

    @instanceoptionalmethod
    @ensure_error_type(JSONValidationError, "Error validating JSON-format string {json_string}: {0}")
    def validate_json_string(self, json_string: str):
        """
        Validates the JSON-format string using this object's validation methods
        (schema, special).

        :param self:            The validator instance/class.
        :param json_string:     The JSON-format string to validate.
        """
        # Convert the string to raw JSON
        raw_json = json.loads(json_string)

        # Validate the raw JSON
        self.validate_raw_json(raw_json)

    @instanceoptionalmethod
    def is_valid_json_string(self, json_string: str) -> bool:
        """
        Checks if the JSON-format string is valid.

        :param self:            The validator instance/class.
        :param json_string:     The JSON-format string to check.
        :return:                True if the JSON is valid,
                                False if not.
        """
        return does_not_raise(self.validate_json_string, json_string)

    @instanceoptionalmethod
    @ensure_error_type(JSONValidationError, "Error getting JSON validation schema: {0}")
    def get_json_validation_schema(self) -> JSONSchema:
        """
        Gets the schema to validate JSON for objects of this type.
        This may be a cached object, so if it is necessary to modify
        the schema, use clone_json_validation_schema instead.

        :param self:    The validator instance/class.
        :return:        The schema.
        """
        if instanceoptionalmethod.will_fail_on_missing_instance(self._get_json_validation_schema):
            raise JSONValidationError(f"Method '{JSONValidator._get_json_validation_schema.__name__}' was "
                                      f"overridden by an instance-method, but was called from the "
                                      f"'{self.__name__}' class")

        return self._get_json_validation_schema()

    @abstractmethod
    @instanceoptionalmethod
    def _get_json_validation_schema(self) -> JSONSchema:
        """
        Gets the schema to validate JSON for objects of this type.
        This may be a cached object, so if it is necessary to modify
        the schema, use clone_json_validation_schema instead.

        :param self:    The validator instance/class.
        :return:        The schema.
        """
        pass

    @instanceoptionalmethod
    @ensure_error_type(JSONValidationError, "Error getting JSON validator: {0}")
    def get_validator(self):
        """
        Gets the jsonschema validator to use to perform JSON validation.

        :param self:    The validator instance/class.
        :return:        The jsonschema validator.
        """
        # Get our schema
        schema: JSONSchema = self.get_json_validation_schema()

        # Get the validator class
        validator_type = jsonschema.validators.validator_for(schema)

        # Check the schema is valid
        validator_type.check_schema(schema)

        # Create the instance
        return validator_type(schema)

    @instanceoptionalmethod
    @ensure_error_type(JSONValidationError, "Error cloning JSON validator: {0}")
    def clone_json_validation_schema(self) -> JSONSchema:
        """
        Gets a copy of the validation schema for this validator.
        Use this if you need to modify the schema.

        :param self:    The validator instance/class.
        :return:        The copy of the schema.
        """
        return deep_copy(self.get_json_validation_schema())

    @instanceoptionalmethod
    @ensure_error_type(JSONValidationError, "Error performing special JSON validation: {0}")
    def perform_special_json_validation(self, raw_json: RawJSONElement):
        """
        Optional method to perform custom validation of raw JSON elements
        for objects of this type. Should throw an exception if the raw JSON
        element doesn't meet the special validation criteria. By default
        does nothing.

        :param self:        The validator instance/class.
        :param raw_json:    The raw JSON to validate.
        """
        self._perform_special_json_validation(raw_json)

    @instanceoptionalmethod
    def _perform_special_json_validation(self, raw_json: RawJSONElement):
        """
        Optional method to perform custom validation of raw JSON elements
        for objects of this type. Should throw an exception if the raw JSON
        element doesn't meet the special validation criteria. By default
        does nothing.

        :param self:        The validator instance/class.
        :param raw_json:    The raw JSON to validate.
        """
        pass



from abc import ABC

from wai.common.decorator import ensure_error_type

from ..error import JSONSerialisationError
from ._JSONSerialisable import JSONSerialisable
from ._JSONDeserialisable import JSONDeserialisable, SelfType


class JSONBiserialisable(JSONSerialisable, JSONDeserialisable[SelfType], ABC):
    """
    Interface for classes which implement both JSONSerialisable and
    JSONDeserialisable[T].
    """
    @ensure_error_type(JSONSerialisationError, "Error copying object using JSON serialisation: {0}")
    def json_copy(self, validate: bool = True) -> SelfType:
        """
        Creates a copy of this object by serialising to JSON and
        then deserialising again.

        :param validate:    Whether to validate the serialised JSON.
        :return:            The copy of this object.
        """
        # Only validate in one direction
        return self.from_raw_json(self.to_raw_json(validate), False)



"""
Package for proxy objects. Proxies are objects which act like other
objects, but can be converted to and from JSON, and subscribe to a
schema which is enforced during programmatic use as well as conversion.
"""
from ._ArrayProxy import ArrayProxy
from ._MapProxy import MapProxy



from abc import abstractmethod, ABC
from typing import Tuple, List, Iterable, Callable, Optional, Any

from ...error import JSONPropertyError, OptionalDisallowed, OfPropertySelectionError
from ...schema import JSONSchema
from .._typing import Absent, PropertyValueType, OptionallyPresent
from ._Property import Property


class OfProperty(Property, ABC):
    """
    Base class for OneOfProperty and AnyOfProperty.
    """
    def __init__(self,
                 name: Optional[str] = None,
                 sub_properties: Iterable[Property] = tuple(),
                 *,
                 schema_function: Callable[[Iterable[JSONSchema]], JSONSchema] = None,  # one_of/any_of/all_of
                 optional: bool = False,
                 default: OptionallyPresent[PropertyValueType] = Absent):
        # Consume the sub-properties
        sub_properties = tuple(sub_properties)

        # Must provide at least 2 sub-properties
        if len(sub_properties) < 2:
            raise JSONPropertyError(f"Of properties require at least 2 sub-properties")

        # Sub-properties can't be optional
        for sub_property in sub_properties:
            if sub_property.is_optional:
                raise OptionalDisallowed(f"Can't use optional properties as sub-properties to any of the"
                                         f"'of' property variations")

        self._sub_properties: Tuple[Property] = sub_properties
        self._schema_function = schema_function

        super().__init__(
            name,
            optional=optional,
            default=default
        )

    def _get_json_validation_schema(self) -> JSONSchema:
        return self._schema_function(
            *(
                prop.get_json_validation_schema()
                for prop in self._sub_properties
            )
        )

    def _validate_value(self, value: Any) -> PropertyValueType:
        # Find the properties this value is valid for
        values = []
        for prop in self._sub_properties:
            # Record the validated value, or Absent if validation failed
            try:
                validated_value = prop.validate_value(value)
            except Exception:
                validated_value = Absent

            values.append(validated_value)

        # Select the canonical index
        subproperty_selection = self._choose_subproperty([value is not Absent for value in values])

        # Make sure the selection is in range
        if not isinstance(subproperty_selection, int) or not (0 <= subproperty_selection < len(self._sub_properties)):
            raise OfPropertySelectionError(f"Error in internal index for Of property. "
                                           f"Index should be integer in [0:{len(self._sub_properties)}), "
                                           f"got {subproperty_selection}")

        # Get the selection
        value = values[subproperty_selection]

        # Make sure the selected value is one of the successful ones
        if value is Absent:
            raise OfPropertySelectionError(f"Error selecting sub-property: selected sub-property "
                                           f"{subproperty_selection} which failed validation")

        return value

    @classmethod
    @abstractmethod
    def _choose_subproperty(cls, successes: List[bool]) -> int:
        """
        Allows the sub-classes to choose which sub-property should
        store a value based on the success or failure of all sub-
        properties validating the value. Should raise an exception if it
        can't decide.

        :param successes:   The list of property validation successes/failures.
        :return:            The index of the property to regard as storing the
                            value.
        """
        pass



"""
Package for working with JSON-format configuration files.
"""
from ._JSONObject import JSONObject
from ._StrictJSONObject import StrictJSONObject
from ._typing import (
    Absent,
    OptionallyPresent,
    PropertyValueType
)



from enum import Enum, auto
from typing import Dict, TypeVar, List, Any, Union, Optional, Iterator, Tuple

from wai.common.meta import instanceoptionalmethod
from wai.common.meta.dynamic_defaults import with_dynamic_defaults, dynamic_default

from ..error import JSONValidationError, RequiredDisallowed, JSONPropertyError
from ..raw import RawJSONElement, RawJSONObject
from ..schema import JSONSchema, standard_object, IS_JSON_SCHEMA, IS_JSON_DEFINITION, TRIVIALLY_FAIL_SCHEMA, is_schema
from ..schema.constants import DEFINITIONS_KEYWORD
from ..serialise import JSONValidatedBiserialisable
from ..validator import StaticJSONValidator
from .property import RawProperty, Property, JSONObjectProperty
from ._typing import Absent, OptionallyPresent, PropertyValueType

# The type of this configuration
SelfType = TypeVar("SelfType", bound="Configuration")

# A schema which accepts any valid JSON value
DEFAULT_SCHEMA: JSONSchema = {DEFINITIONS_KEYWORD: IS_JSON_DEFINITION}
DEFAULT_SCHEMA.update(IS_JSON_SCHEMA)


def additional_properties_validation_as_property(validation: Union[JSONSchema, Property, None]) -> Optional[Property]:
    """
    Helper function which converts the prescribed additional properties validation
    into a property.

    :param validation:  The validation as returned from JSONObject._additional_properties_validation.
    :return:            The additional properties property.
    """
    # Return None for no validation or trivially-failing validation
    if validation is None or validation is TRIVIALLY_FAIL_SCHEMA:
        return None

    # If it's already a property, just make sure it's optional
    if isinstance(validation, Property):
        if not validation.is_optional:
            raise RequiredDisallowed(f"Validation for additional properties must be optional")

        return validation

    # Otherwise it's a schema, so make sure it's valid
    if not is_schema(validation):
        raise JSONValidationError(f"Validation for additional properties must be a property or a schema, "
                                  f"not a {type(validation).__name__}: {validation}")

    return RawProperty(schema=validation, optional=True)


class PropertyOptionality(Enum):
    """
    Enumeration of the types of optionality a property can have.
    """
    REQUIRED = auto()
    OPTIONAL = auto()
    ADDITIONAL = auto()


class JSONObject(JSONValidatedBiserialisable[SelfType], StaticJSONValidator):
    """
    Base class for Python representations of JSON-format configuration files.
    All configurations are validated against a schema for format correctness.
    The attributes of a configuration should be either JSON types or nested
    configurations.
    """
    # Forward-declarations of class attributes
    # If JSONObject is instantiated directly, it has only additional properties
    # and they can be anything
    _required_properties: Dict[str, Property] = {}
    _optional_properties: Dict[str, Property] = {}
    _additional_property: Optional[Property] = additional_properties_validation_as_property(DEFAULT_SCHEMA)

    def __init__(self, **initial_values):
        # Create the property values container
        self._property_values: Dict[str, PropertyValueType] = {}

        # Attempt to convert attribute names to property names
        self._attribute_to_property_names(initial_values)

        # Make sure all required properties have values
        for required_property in self._required_properties:
            if required_property not in initial_values:
                raise JSONPropertyError(f"Value for required property '{required_property}' not set")

        # Set all properties named in initial values
        for name, value in initial_values.items():
            self.set_property(name, value)

    @instanceoptionalmethod
    def has_property(self, name: str, *,
                     require_value: bool = False) -> bool:
        """
        Whether this object has a property with the given name. If the object
        allows additional properties, this will include additional properties
        which currently have values, and required/optional properties regardless
        of value.

        :param name:            The property name.
        :param require_value:   Whether properties without values should be regarded
                                as valid. Only available on instances.
        :return:                True if the object has the property.
        """
        # If called on the class, we can't require a value
        if not instanceoptionalmethod.is_instance(self):
            if require_value:
                raise JSONPropertyError(f"Can't require a value without an instance")

        # If called on an instance, check for values
        else:
            # If it has a value, it's always present
            if name in self._property_values:
                return True

            # If a value is required but missing, it's not considered present
            if require_value:
                return False

        # In all other cases, the property must be defined
        return name in self._required_properties or name in self._optional_properties

    def get_property(self, name: str, *,
                     bypass_default: bool = False) -> OptionallyPresent[PropertyValueType]:
        """
        Gets the value of a property by name.

        :param name:            The name of the property.
        :param bypass_default:  Whether to return Absent instead of the default value.
        :return:                The property's value.
        """
        # Get the property's value if we have one
        if name in self._property_values:
            return self._property_values[name]

        # If we don't want the default value, return Absent
        if bypass_default:
            return Absent

        return self._get_property(name).default

    def get_property_as_raw_json(self, name: str, *,
                                 bypass_default: bool = False) -> OptionallyPresent[RawJSONElement]:
        """
        Gets the value of a property as raw JSON.

        :param name:            The property name.
        :param bypass_default:  Whether to return Absent instead of the default value.
        :return:                The raw JSON, or Absent if the property has no value.
        """
        # Avoid unnecessary copying if the default is a serialisable
        if name not in self._property_values:
            return self._get_property(name).default_as_raw_json

        # Get the value
        value = self.get_property(name, bypass_default=bypass_default)

        # If it's a serialisable type, serialise it
        if isinstance(value, JSONValidatedBiserialisable):
            value = value.to_raw_json()

        return value

    def set_property(self, name: str, value: OptionallyPresent[PropertyValueType]):
        """
        Sets the value of a property by name.

        :param name:    The property to set.
        :param value:   The value to set.
        """
        # Get the property for setting the value
        prop = self._get_property(name)

        # Use the property to validate the value
        value = prop.validate_value(value)

        if value is Absent:
            if name in self._property_values:
                del self._property_values[name]
        else:
            self._property_values[name] = value

    def delete_property(self, name: str):
        """
        Removes the given property's value from this object.

        :param name:    The property's name.
        """
        self.set_property(name, Absent)

    @classmethod
    def get_property_optionality(cls, name: str) -> PropertyOptionality:
        """
        Gets the optionality of the named property.

        :param name:    The property name.
        :return:        The property's optionality.
        """
        # Get the property
        prop = cls._get_property(name)

        return (PropertyOptionality.ADDITIONAL if prop is cls._additional_property else
                PropertyOptionality.OPTIONAL if prop.is_optional else
                PropertyOptionality.REQUIRED)

    @instanceoptionalmethod
    def properties(self) -> Iterator[str]:
        """
        Iterates through all properties of the object, including additional
        properties if called on the instance.

        :return:    An iterator over the property names.
        """
        # Required properties first
        for name in self._required_properties:
            yield name

        # Optional properties
        for name in self._optional_properties:
            yield name

        # Additional properties (if called on the instance)
        if instanceoptionalmethod.is_instance(self):
            for name in self._property_values:
                if name not in self._required_properties and name not in self._optional_properties:
                    yield name

    def values(self, *,
               bypass_default: bool = False,
               include_absent: bool = False) -> Iterator[OptionallyPresent[PropertyValueType]]:
        """
        Iterates over the values of all properties in the object.

        :param bypass_default:  Whether to bypass default values
        :param include_absent:  Whether to include absent values.
        :return:                An iterator over the property values.
        """
        return (value for name, value in self.items(bypass_default=bypass_default,
                                                    include_absent=include_absent))

    def items(self, *,
              bypass_default: bool = False,
              include_absent: bool = False) -> Iterator[Tuple[str, OptionallyPresent[PropertyValueType]]]:
        """
        Iterates through all properties in this object.

        :param bypass_default:  Whether to bypass default values
        :param include_absent:  Whether to include absent values.
        :return:                An iterator of name, value pairs for the properties.
        """
        for name in self.properties():
            value = self.get_property(name, bypass_default=bypass_default)
            if include_absent or value is not Absent:
                yield name, value

    @classmethod
    def allows_additional_properties(cls) -> bool:
        """
        Whether this object can contain additional properties.

        :return:    True if the object can contain additional properties.
        """
        return cls._additional_property is not None

    @classmethod
    def as_property(cls, name: Optional[str] = None, *, optional: bool = False) -> JSONObjectProperty:
        """
        Creates a property (e.g. for another JSON object) which takes instances
        of this type of JSON object as values.

        :param name:        The name to give the property (default is to inherit the attribute name).
        :param optional:    Whether the property should be optional.
        :return:            The property.
        """
        return JSONObjectProperty(name, cls, optional=optional)

    @classmethod
    def _get_property(cls, name: str) -> Property:
        """
        Gets the property descriptor to handle the named property.

        :param name:    The property name.
        :return:        The property descriptor.
        """
        if name in cls._required_properties:
            return cls._required_properties[name]
        elif name in cls._optional_properties:
            return cls._optional_properties[name]
        elif cls.allows_additional_properties():
            return cls._additional_property

        raise JSONPropertyError(f"JSONObject '{cls.__name__}' has no property '{name}', "
                                f"and doesn't allow additional properties")

    def __getitem__(self, item: str) -> OptionallyPresent[PropertyValueType]:
        return self.get_property(item)

    def __setitem__(self, key: str, value: OptionallyPresent[PropertyValueType]):
        self.set_property(key, value)

    def __delitem__(self, key: str):
        self.delete_property(key)

    def __contains__(self, item: str) -> bool:
        return self.has_property(item)

    def __iter__(self):
        return self.properties()

    def _serialise_to_raw_json(self) -> RawJSONObject:
        return {name: self.get_property_as_raw_json(name)
                for name in self._property_values}

    @classmethod
    def _deserialise_from_raw_json(cls, raw_json: RawJSONObject) -> SelfType:
        return cls(**raw_json)

    @classmethod
    def _get_json_validation_schema(cls) -> JSONSchema:
        # Extract the required property schemas
        required_properties_schema: Dict[str, JSONSchema] = {
            name: prop.get_json_validation_schema()
            for name, prop in cls._required_properties.items()
        }

        # Extract the optional property schemas
        optional_properties_schema: Dict[str, JSONSchema] = {
            name: prop.get_json_validation_schema()
            for name, prop in cls._optional_properties.items()
        }

        # If no additional properties allowed, return a schema indicating this
        if not cls.allows_additional_properties():
            return standard_object(required_properties_schema, optional_properties_schema)

        # Extract the additional properties schema
        additional_properties_schema: JSONSchema = cls._additional_property.get_json_validation_schema()

        # Create the schema
        return standard_object(required_properties_schema,
                               optional_properties_schema,
                               additional_properties=additional_properties_schema)

    @classmethod
    def _get_all_defined_properties(cls) -> Dict[str, Property]:
        """
        Gets all defined properties of this configuration.

        :return:    A dictionary of the properties of this configuration.
        """
        # Get all attribute names
        attribute_names: List[str] = dir(cls)

        # Get all public attributes
        attributes: List[Any] = [getattr(cls, name)
                                 for name in attribute_names
                                 if not name.startswith("_")]

        # Create the empty properties dict
        defined_properties = {}

        # Add each property in turn, checking for duplicate names
        for attribute in attributes:
            # Skip anything that's not a property
            if not isinstance(attribute, Property):
                continue

            # Check if another property is already using the name
            if attribute.name in defined_properties:
                raise JSONPropertyError(f"Multiple properties named '{attribute.name}' "
                                        f"for JSON object '{cls.__qualname__}'")

            # Add the property
            defined_properties[attribute.name] = attribute

        return defined_properties

    @classmethod
    def _attribute_to_property_names(cls, values: Dict[str, Any]):
        """
        Tries to convert any initial values specified by attribute name
        into property names instead. Modifies the dictionary in-place.

        :param values:  The initial values to the constructor.
        """
        # Create a mapping from attribute names to property names
        mapping = {}
        for key, value in values.items():
            # Skip defined properties
            if key in cls._required_properties or key in cls._optional_properties:
                continue

            # Try and get the attribute the name refers to (if it does so)
            try:
                attr = getattr(cls, key, None)

            # If there is no attribute with that name, leave it as an additional property
            except AttributeError:
                pass

            # Otherwise see if the attribute is a property, and if so, record the mapping
            else:
                if isinstance(attr, Property):
                    mapping[key] = attr.name

        # Implement the mapping
        for attribute_name, property_name in mapping.items():
            # If the property name is already in the values, it's doubly-defined
            if property_name in values:
                raise JSONPropertyError(f"Property '{property_name}' was specified twice in "
                                        f"initial values, once by its property name, and once "
                                        f"by its attribute name '{attribute_name}'")

            values[property_name] = values[attribute_name]
            del values[attribute_name]

    @classmethod
    def _additional_properties_validation(cls) -> Union[JSONSchema, Property, None]:
        """
        Gets the schema/property to validate any additional properties set
        on this configuration. By default all additional properties are allowed.
        Override to modify this behaviour.
        """
        return DEFAULT_SCHEMA

    @with_dynamic_defaults
    def __init_subclass__(cls,
                          programmatic_properties: Dict[str, Property] = dynamic_default(dict),
                          **kwargs):
        # Perform any super initialisation
        super().__init_subclass__(**kwargs)

        # Hack for adding programmatic properties, as Generic meta-class
        # stops __set_name__ from being called
        for name, prop in programmatic_properties.items():
            setattr(cls, name, prop)
            prop.__set_name__(cls, name)

        # Get all of the defined properties
        defined_properties = cls._get_all_defined_properties()

        # Filter the required properties
        cls._required_properties: Dict[str, Property] = {
            name: value
            for name, value in defined_properties.items()
            if not value.is_optional
        }

        # Filter the optional properties
        cls._optional_properties: Dict[str, Property] = {
            name: value
            for name, value in defined_properties.items()
            if value.is_optional
        }

        # Create the additional property if allowed
        cls._additional_property: Optional[Property] = additional_properties_validation_as_property(
            cls._additional_properties_validation()
        )



from abc import ABC
from typing import Type, Optional, Any

from wai.common.abc import is_abstract_class

from ...error import JSONPropertyError
from ...schema import JSONSchema
from ...serialise import JSONValidatedBiserialisable
from .._typing import PropertyValueType, Absent, OptionallyPresent
from ._Property import Property


class ProxyProperty(Property, ABC):
    """
    Property which takes values of a type that is serialisable
    to/from JSON.
    """
    def __init__(self,
                 name: Optional[str] = None,
                 *,
                 proxy: Type[JSONValidatedBiserialisable] = JSONValidatedBiserialisable,
                 optional: bool = False,
                 default: OptionallyPresent[PropertyValueType] = Absent):
        # Type argument must be concrete
        if is_abstract_class(proxy):
            raise JSONPropertyError(f"Proxy type argument must be a concrete class, got {proxy.__name__}")

        # The type of values this property takes
        self._type: Type[JSONValidatedBiserialisable] = proxy

        super().__init__(
            name,
            optional=optional,
            default=default
        )

    @property
    def proxy_type(self) -> Type[JSONValidatedBiserialisable]:
        """
        Gets the type of values this proxy-property takes.
        """
        return self._type

    def _get_json_validation_schema(self) -> JSONSchema:
        # Use the value-type's schema
        return self._type.get_json_validation_schema()

    def _validate_value(self, value: Any) -> PropertyValueType:
        # Must be an instance of the correct type, or JSON deserialisable to the correct type
        if not isinstance(value, self._type):
            try:
                if isinstance(value, str):
                    return self._type.from_json_string(value)
                else:
                    return self._type.from_raw_json(value)
            except Exception as e:
                raise JSONPropertyError(f"Error validating proxy value: {e}") from e

        return value



"""
Package specifying interfaces for objects/classes which can perform JSON validation.
"""
from ._BasicSchemaValidator import BasicSchemaValidator
from ._JSONValidator import JSONValidator
from ._StaticJSONValidator import StaticJSONValidator



from typing import Optional, Any

from .._typing import PropertyValueType, Absent, OptionallyPresent
from .proxies import MapProxy
from ._Property import Property
from ._ProxyProperty import ProxyProperty
from ._RawProperty import RawProperty


class MapProperty(ProxyProperty):
    """
    Property which validates a map of strings to some other type.
    """
    def __init__(self,
                 name: Optional[str] = None,
                 value_property: Property = RawProperty(),
                 *,
                 optional: bool = False,
                 default: OptionallyPresent[PropertyValueType] = Absent):
        super().__init__(
            name,
            proxy=MapProxy.specify(value_property),
            optional=optional,
            default=default
        )

    def _validate_value(self, value: Any) -> PropertyValueType:
        # Convert other map-proxies and raw dictionaries to our proxy-type
        if ((isinstance(value, MapProxy) and not isinstance(value, self.proxy_type))
                or isinstance(value, dict)):
            return self.proxy_type(value)

        return super()._validate_value(value)



from abc import abstractmethod, ABC
from typing import Iterable, Optional, Dict, Union, Mapping, Type

from ....error import OptionalDisallowed
from ....serialise import JSONValidatedBiserialisable
from ....schema import JSONSchema, standard_object
from ....validator import StaticJSONValidator
from ..._typing import RawJSONElement, PropertyValueType
from .._Property import Property


class MapProxy(StaticJSONValidator, JSONValidatedBiserialisable['MapProxy'], ABC):
    """
    Class which acts like a map, but validates its elements using a property.
    """
    def __init__(self,
                 initial_values: Optional[Union[Iterable, Mapping]] = None,
                 **kwargs):
        # The values in the map
        self._values: Dict[str, PropertyValueType] = {}

        # Add any initial values
        self.update(initial_values, **kwargs)

    @staticmethod
    @abstractmethod
    def value_property() -> Property:
        """
        Gets the property which is used to validate elements of the map.
        """
        pass

    @staticmethod
    def specify(value_property: Property) -> Type['MapProxy']:
        """
        Creates a specific map proxy-type.

        :param value_property:      The property to use to validate map values.
        :return:                    The map proxy-type matching the specification.
        """
        # Create the proxy-type as a closure
        class ClosureMapProxy(MapProxy):
            @staticmethod
            def value_property() -> Property:
                return value_property

        return ClosureMapProxy

    def _serialise_to_raw_json(self) -> RawJSONElement:
        return {key: value.to_raw_json() if isinstance(value, JSONValidatedBiserialisable) else value
                for key, value in self._values.items()}

    @classmethod
    def _deserialise_from_raw_json(cls, raw_json: RawJSONElement) -> 'MapProxy':
        return cls(raw_json)

    @classmethod
    def _get_json_validation_schema(cls) -> JSONSchema:
        return standard_object(additional_properties=cls.value_property().get_json_validation_schema())

    def __init_subclass__(cls, **kwargs):
        # Make sure the sub-property isn't optional
        if cls.value_property().is_optional:
            raise OptionalDisallowed("Can't use optional sub-property with maps")

        super().__init_subclass__(**kwargs)

    # ------------ #
    # DICT METHODS #
    # ------------ #

    def clear(self):
        self._values.clear()

    def copy(self):
        return type(self)(self)

    @staticmethod
    def fromkeys(seq, value):
        # TODO: Implement
        raise NotImplementedError(MapProxy.fromkeys.__qualname__)

    def get(self, k: str, default=None):
        return self._values.get(k, default)

    def items(self):
        return self._values.items()

    def keys(self):
        return self._values.keys()

    def pop(self, k, d=None):
        return self._values.pop(k, d)

    def popitem(self):
        return self._values.popitem()

    def setdefault(self, key, default):
        if key not in self:
            self[key] = default

        return self[key]

    def update(self, E=None, **F):
        if E is not None:
            if hasattr(E, "keys") and callable(E.keys):
                for k in E:
                    self[k] = E[k]
            else:
                for k, v in E:
                    self[k] = v

        for k in F:
            self[k] = F[k]

    def values(self):
        return self._values.values()

    def __contains__(self, key: str) -> bool:
        return key in self._values

    def __delitem__(self, key: str):
        del self._values[key]

    def __eq__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(MapProxy.__eq__.__qualname__)

    def __getitem__(self, y):
        return self._values[y]

    def __ge__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(MapProxy.__ge__.__qualname__)

    def __gt__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(MapProxy.__gt__.__qualname__)

    def __iter__(self):
        return iter(self._values)

    def __len__(self):
        return len(self._values)

    def __le__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(MapProxy.__le__.__qualname__)

    def __lt__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(MapProxy.__lt__.__qualname__)

    def __ne__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(MapProxy.__ne__.__qualname__)

    def __repr__(self):
        # TODO: Implement
        raise NotImplementedError(MapProxy.__repr__.__qualname__)

    def __setitem__(self, key: str, value):
        self._values[key] = self.value_property().validate_value(value)

    def __str__(self):
        return str(self._values)



"""
Module for type related to JSON objects.
"""
from typing import TypeVar, Union, Optional

from ..serialise import JSONValidatedBiserialisable
from ..raw import RawJSONElement

# ====== #
# ABSENT #
# ====== #


class AbsentType:
    """
    Singleton class representing a property that is absent from
    a configuration. This is used instead of None, because None
    is a valid value for some properties (represents JSON null).
    """
    # The singleton instance
    __instance: Optional['AbsentType'] = None

    # Create the instance the first time, then just return it
    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)

        return cls.__instance

    def __str__(self) -> str:
        return "Absent"


# Create the singleton instance for export
Absent = AbsentType()


# ================= #
# OptionallyPresent #
# ================= #


# The type of the value if it is present
PresentType = TypeVar("PresentType")

OptionallyPresent = Union[AbsentType, PresentType]
"""
Type which can take values of some other type or the singleton Absent value.

E.g. a: OptionallyPresent[int] = |any int or Absent|
"""


# ================= #
# PropertyValueType #
# ================= #


# The type of values a property can take
PropertyValueType = Union[RawJSONElement, JSONValidatedBiserialisable]



from ._JSONSchemaError import JSONSchemaError


class SchemaDefinitionRedefined(JSONSchemaError):
    """
    Error for when consolidating schema definitions and
    an attempt is made to redefine a definition.
    """
    def __init__(self, definition_name: str, original_definition, new_definition):
        super().__init__(f"Attempted to redefine JSON schema definition '{definition_name}': "
                         f"from {original_definition} to {new_definition}")



from abc import ABC, abstractmethod
from typing import Any, Optional

from ...error import JSONPropertyError
from ...raw import deep_copy, RawJSONElement
from ...serialise import JSONValidatedBiserialisable
from ...validator import StaticJSONValidator
from .._typing import PropertyValueType, Absent, OptionallyPresent


class Property(StaticJSONValidator, ABC):
    """
    A property of a configuration describes a key to the object and the types
    of value it can take. The value of a property is a raw JSON element, or something
    that can be converted to/from raw JSON. The property name can be set to the
    empty string to inherit the attribute name it is set against in the
    configuration.
    """
    def __init__(self,
                 name: Optional[str] = None,  # Default tells the property to inherit its attribute name
                 *,
                 optional: bool = False,
                 default: OptionallyPresent[PropertyValueType] = Absent):
        # Can't use the empty string as a name
        if name == "":
            raise JSONPropertyError("Can't name properties with the empty string")

        self._name: str = name if name is not None else ""
        self._attribute_name: str = ""
        self._optional: bool = optional

        # Validate the default value
        validated_default = Absent
        if optional:
            validated_default = self.validate_value(default)

        # Can only change the default value if the property is optional
        elif default is not Absent:
            raise JSONPropertyError("Can't set default values for required properties")

        # Save a copy of the default value, unless it is absent or
        # validation already created a new object
        self._default: OptionallyPresent[PropertyValueType] = (
            Absent if default is Absent else
            validated_default if validated_default is not default else
            validated_default.json_copy(False) if isinstance(validated_default, JSONValidatedBiserialisable) else
            deep_copy(validated_default)
        )

    @property
    def name(self) -> str:
        """
        Gets the name of this property.

        :return:    The property name.
        """
        return self._name

    @property
    def is_optional(self) -> bool:
        """
        Whether this property is an optional property.
        """
        return self._optional

    @property
    def default(self) -> OptionallyPresent[PropertyValueType]:
        """
        Gets the default value for this property.
        """
        # Can't get the default of a required property
        if not self.is_optional:
            raise AttributeError(f"Required property '{self.name}' has no default value")

        if self._default is Absent:
            return Absent
        elif isinstance(self._default, JSONValidatedBiserialisable):
            return self._default.json_copy(False)
        else:
            return deep_copy(self._default)

    @property
    def default_as_raw_json(self) -> OptionallyPresent[RawJSONElement]:
        """
        Gets the default value as raw JSON.
        """
        if isinstance(self._default, JSONValidatedBiserialisable):
            return self._default.to_raw_json(False)
        else:
            return self.default

    @property
    def has_default(self) -> bool:
        """
        Whether this property has a default value set.
        """
        return self._default is not Absent

    def __get__(self, instance, owner):
        # If accessed from the class, return the property itself
        if instance is None:
            return self

        # Check the instance is valid
        self._check_instance(instance, "get")

        return instance.get_property(self._name)

    def __set__(self, instance, value):
        # Check the instance is valid
        self._check_instance(instance, "set")

        # Don't need to validate the value as JSONObject will
        # call back to us as part of set_property.
        instance.set_property(self._name, value)

    def __delete__(self, instance):
        # Check the instance is valid
        self._check_instance(instance, "delete")

        instance.delete_property(self._name)

    def __set_name__(self, owner, name: str):
        # Can't reset the name
        if self._attribute_name != "" and self._attribute_name != name:
            raise JSONPropertyError(f"Attempted to reassign property '{self._name}' to attribute '{name}' "
                                    f"(already assigned to attribute '{self._attribute_name}')")

        # Can't have private properties (except for _additional_property)
        if name.startswith("_") and name != "_additional_property":
            raise JSONPropertyError(f"Property attribute names can't start with underscores ({name})")

        # Save the attribute name
        self._attribute_name = name

        # Inherit the attribute name as the property name if one wasn't
        # explicitly set
        if self._name == "":
            self._name = name

    @staticmethod
    def _check_instance(instance: Any, method: str):
        """
        Checks that the instance is available and a JSON object.

        :param instance:    The instance to check.
        :param method:      The method checking the instance, for error messages.
        """
        # JSONObject requires local import to avoid circular dependency
        from .._JSONObject import JSONObject

        # Instance must be a JSON object
        if not isinstance(instance, JSONObject):
            raise JSONPropertyError(f"Can only {method} value from a property "
                                    f"if called from a JSON object")

    def validate_value(self, value: Any) -> OptionallyPresent[PropertyValueType]:
        """
        Performs property value validation. Should raise an exception if
        validation fails, or return the actual value to store if validation
        passes.

        :param value:       The value to validate.
        :return:            The value to store.
        """
        # Default checking is for absent values on required properties
        if value is Absent:
            if not self.is_optional:
                raise JSONPropertyError(f"Required property '{self.name}' can't be set absent")

            return value

        return self._validate_value(value)

    @abstractmethod
    def _validate_value(self, value: Any) -> PropertyValueType:
        """
        Performs property value validation. Should raise an exception if
        validation fails, or return the actual value to store if validation
        passes.

        :param value:       The value to validate.
        :return:            The value to store.
        """
        pass



from typing import Optional

from ...schema import BOOL_SCHEMA
from .._typing import PropertyValueType, Absent, OptionallyPresent
from ._RawProperty import RawProperty


class BoolProperty(RawProperty):
    """
    Configuration property which validates a boolean value.
    """
    def __init__(self,
                 name: Optional[str] = None,
                 *,
                 optional: bool = False,
                 default: OptionallyPresent[PropertyValueType] = Absent):
        super().__init__(
            name,
            schema=BOOL_SCHEMA,
            optional=optional,
            default=default
        )



from typing import Iterable, Optional, FrozenSet

from ...raw import RawJSONPrimitive
from ...schema import enum
from .._typing import PropertyValueType, Absent, OptionallyPresent
from ._RawProperty import RawProperty


class EnumProperty(RawProperty):
    """
    Property which accepts one of a set of specific primitive values.
    """
    def __init__(self,
                 name: Optional[str] = None,
                 *,
                 values: Iterable[RawJSONPrimitive] = tuple(),
                 optional: bool = False,
                 default: OptionallyPresent[PropertyValueType] = Absent):
        # Consume the iterable
        self._values: FrozenSet[RawJSONPrimitive] = frozenset(values)

        super().__init__(
            name,
            schema=enum(*self._values),
            optional=optional,
            default=default
        )

    @property
    def values(self) -> FrozenSet[RawJSONPrimitive]:
        """
        Gets the values this property takes.

        :return:    The set of values.
        """
        return self._values



"""
Package for specialised error types to do with processing JSON.
"""
from ._JSONError import JSONError
from ._JSONPropertyError import JSONPropertyError
from ._JSONSchemaError import JSONSchemaError
from ._JSONSerialisationError import JSONSerialisationError
from ._JSONValidationError import JSONValidationError
from ._OfPropertySelectionError import OfPropertySelectionError
from ._OptionalDisallowed import OptionalDisallowed
from ._RequiredDisallowed import RequiredDisallowed
from ._SchemaDefinitionRedefined import SchemaDefinitionRedefined



"""
Package for working with "raw" JSON. Raw JSON is JSON represented
as Python primitives, as defined by the built-in Python "json" module.

The mapping from JSON types to Python types is as follows:

JSON                Python
====                ======
Object      <->     dict
Array       <->     list, tuple
String      <->     str
Number      <->     int, float
Boolean     <->     bool
Null        <->     None
"""
from ._deep_copy import deep_copy
from ._dynamic import (
    is_raw_json_object,
    is_raw_json_array,
    is_raw_json_string,
    is_raw_json_number,
    is_raw_json_bool,
    is_raw_json_null,
    is_raw_json_primitive,
    is_raw_json_element,
    is_raw_json_element_type
)
from ._typing import (
    RawJSONObject,
    RawJSONArray,
    RawJSONString,
    RawJSONNumber,
    RawJSONBool,
    RawJSONNull,
    RawJSONPrimitive,
    RawJSONElement
)



from typing import List, Iterable, Optional

from ...error import OfPropertySelectionError
from ...schema import all_of
from .._typing import PropertyValueType, Absent, OptionallyPresent
from ._OfProperty import OfProperty
from ._Property import Property


class AllOfProperty(OfProperty):
    """
    Property which validates JSON that matches all of
    a number of sub-properties.
    """
    def __init__(self,
                 name: Optional[str] = None,
                 sub_properties: Iterable[Property] = tuple(),
                 *,
                 optional: bool = False,
                 default: OptionallyPresent[PropertyValueType] = Absent):
        super().__init__(
            name,
            sub_properties,
            schema_function=all_of,
            optional=optional,
            default=default
        )

    @classmethod
    def _choose_subproperty(cls, successes: List[bool]) -> int:
        # If all sub-properties match, just pick the first one
        if all(successes):
            return 0

        raise OfPropertySelectionError(f"Value didn't match all sub-properties")



from typing import Optional, Any

from .._typing import PropertyValueType, Absent, OptionallyPresent
from .proxies import ArrayProxy
from ._Property import Property
from ._ProxyProperty import ProxyProperty
from ._RawProperty import RawProperty


class ArrayProperty(ProxyProperty):
    """
    Property which validates a regular array of elements.
    """
    def __init__(self,
                 name: Optional[str] = None,
                 element_property: Property = RawProperty(),
                 *,
                 min_elements: int = 0,
                 max_elements: Optional[int] = None,
                 unique_elements: bool = False,
                 optional: bool = False,
                 default: OptionallyPresent[PropertyValueType] = Absent):
        super().__init__(
            name,
            proxy=ArrayProxy.specify(element_property,
                                     min_elements,
                                     max_elements,
                                     unique_elements),
            optional=optional,
            default=default
        )

    def _validate_value(self, value: Any) -> PropertyValueType:
        # Convert other proxy-arrays and raw lists/tuples to our proxy-type
        if ((isinstance(value, ArrayProxy) and not isinstance(value, self.proxy_type))
                or isinstance(value, list)
                or isinstance(value, tuple)):
            return self.proxy_type(value)

        return super()._validate_value(value)



from typing import Optional

from ...raw import RawJSONPrimitive
from ...schema import constant
from .._typing import PropertyValueType, Absent, OptionallyPresent
from ._RawProperty import RawProperty


class ConstantProperty(RawProperty):
    """
    Configuration property which validates a constant primitive value.
    """
    def __init__(self,
                 name: Optional[str] = None,
                 value: RawJSONPrimitive = None,
                 *,
                 optional: bool = False,
                 default: OptionallyPresent[PropertyValueType] = Absent):
        self._value = value

        super().__init__(
            name,
            schema=constant(value),
            optional=optional,
            default=default
        )

    @property
    def value(self) -> RawJSONPrimitive:
        """
        Gets the constant value this property takes.

        :return:    The constant value.
        """
        return self._value



from wai.test import AbstractTest
from wai.test.decorators import RegressionTest, Test, ExceptionTest

from json import loads

from wai.json.object import JSONObject
from wai.json.object.property import *


class JSONObjectTest(AbstractTest):
    """
    Unit tests for JSON objects.
    """
    @classmethod
    def subject_type(cls):
        # Default test subject is a plain object
        return JSONObject

    @Test
    def properties(self, subject: JSONObject):
        """
        Test creating a new sub-class of JSONObject.
        """
        class TestObject(JSONObject):
            a = NumberProperty()
            b = StringProperty()
            c = BoolProperty()
            d = EnumProperty(values=(1, 2, "three", "blue"))
            e = ConstantProperty(value="test")
            f = ArrayProperty(element_property=NumberProperty())
            g = MapProperty(value_property=StringProperty())
            h = JSONObjectProperty(object_type=JSONObject)
            i = AnyOfProperty(sub_properties=(NumberProperty(), StringProperty()))
            j = AllOfProperty(sub_properties=(NumberProperty(multiple_of=3), NumberProperty(multiple_of=5)))
            k = OneOfProperty(sub_properties=(NumberProperty(multiple_of=3), NumberProperty(multiple_of=5)))

    @Test
    def set_additional_property(self, subject: JSONObject):
        subject.set_property("test", 13)

        self.assertDictEqual(subject.to_raw_json(), {"test": 13})



"""
Static types for raw JSON objects.
"""
from typing import Dict, Union, List, Tuple

# The type of a raw JSON object in Python-land
RawJSONObject = Dict[str, 'RawJSONElement']

# The type of a raw JSON array in Python-land
RawJSONArray = Union[List['RawJSONElement'], Tuple['RawJSONElement', ...]]

# The type of a raw JSON string in Python-land
RawJSONString = str

# The type of an arbitrary raw JSON number in Python-land
RawJSONNumber = Union[int, float]

# The type of a raw JSON bool in Python-land
RawJSONBool = bool

# The type of a raw JSON null in Python-land
RawJSONNull = None

# The type of a raw JSON primitive
RawJSONPrimitive = Union[RawJSONString, RawJSONNumber, RawJSONBool, RawJSONNull]

# The type of an arbitrary raw JSON element in Python-land
RawJSONElement = \
    Union[
        RawJSONObject,  # Object
        RawJSONArray,  # Array
        RawJSONPrimitive  # Any primitive
    ]



from abc import ABC

from wai.common.meta import instanceoptionalmethod

from ..schema import JSONSchema
from ._JSONValidator import JSONValidator


class StaticJSONValidator(JSONValidator, ABC):
    """
    Base class for JSON validators whose schema is
    constant for the lifetime of the class/object.
    """
    @instanceoptionalmethod
    def get_json_validation_schema(self) -> JSONSchema:
        if not hasattr(self, "__json_validation_schema"):
            setattr(self, "__json_validation_schema", super().get_json_validation_schema())

        return getattr(self, "__json_validation_schema")

    @instanceoptionalmethod
    def get_validator(self):
        if not hasattr(self, "__validator"):
            setattr(self, "__validator", super().get_validator())

        return getattr(self, "__validator")



import json
from abc import abstractmethod
from typing import TypeVar, Generic, IO

from wai.common.decorator import ensure_error_type

from ..error import JSONSerialisationError
from ..raw import RawJSONElement
from ..validator import JSONValidator

# The type of the object that is deserialised
SelfType = TypeVar("SelfType", bound="JSONDeserialisable")


class JSONDeserialisable(Generic[SelfType]):
    """
    Interface for classes which can deserialise instances from JSON.
    """
    @classmethod
    @abstractmethod
    def _deserialise_from_raw_json(cls, raw_json: RawJSONElement) -> SelfType:
        """
        Implements the actual deserialisation from JSON.

        :param raw_json:    The raw JSON representation.
        :return:            An instance of the type.
        """
        pass

    @classmethod
    @ensure_error_type(JSONSerialisationError, "Error deserialising raw JSON {raw_json}: {0}")
    def from_raw_json(cls, raw_json: RawJSONElement, validate: bool = True) -> SelfType:
        """
        Instantiates an object of this type from a raw JSON element.

        :param raw_json:    The raw JSON element.
        :param validate:    Whether to validate the JSON before deserialisation.
        :return:            The object instance.
        """
        # Validate the raw JSON if we are capable
        if validate and issubclass(cls, JSONValidator):
            cls.validate_raw_json(raw_json)

        # Deserialise
        return cls._deserialise_from_raw_json(raw_json)

    @classmethod
    @ensure_error_type(JSONSerialisationError, "Error deserialising JSON string '{json_string}': {0}")
    def from_json_string(cls, json_string: str, validate: bool = True) -> SelfType:
        """
        Instantiates an object of this type from a JSON-format string.

        :param json_string:     The JSON-format string to parse.
        :param validate:        Whether to validate the JSON before deserialisation.
        :return:                The object instance.
        """
        raw_json = json.loads(json_string)

        return cls.from_raw_json(raw_json, validate)

    @classmethod
    @ensure_error_type(JSONSerialisationError, "Error reading JSON from stream: {0}")
    def read_json_from_stream(cls, stream: IO[str], validate: bool = True) -> SelfType:
        """
        Instantiates an object of this type from the given string-stream.

        :param stream:      The stream to read from.
        :param validate:    Whether to validate the JSON before deserialisation.
        :return:            The object instance.
        """
        return cls.from_raw_json(json.load(stream), validate)

    @classmethod
    @ensure_error_type(JSONSerialisationError, "Error loading JSON from file '{filename}': {0}")
    def load_json_from_file(cls, filename: str, validate: bool = True) -> SelfType:
        """
        Loads an instance of this class from the given file.

        :param filename:    The name of the file to load from.
        :param validate:    Whether to validate the JSON before deserialisation.
        :return:            The instance.
        """
        with open(filename, 'r') as file:
            return cls.read_json_from_stream(file, validate)



"""
Package for working with raw JSON schema.
"""
from ._definitions import (
    IS_JSON_DEFINITION,
    IS_JSON_SCHEMA,
    IS_JSON_REFERENCE,
    IS_JSON_REFEREND,
    extract_definitions,
    consolidate_definitions,
    reference_schema,
    reference_string
)
from ._dynamic import (
    string_schema,
    standard_object,
    constant,
    enum,
    number,
    regular_array,
    one_of,
    any_of,
    all_of
)
from ._static import (
    TRIVIALLY_SUCCEED_SCHEMA,
    TRIVIALLY_FAIL_SCHEMA,
    BOOL_SCHEMA,
    INT_SCHEMA,
    FLOAT_SCHEMA,
    STRING_SCHEMA,
    POSITIVE_INTEGER_SCHEMA,
    NEGATIVE_INTEGER_SCHEMA,
    NON_NEGATIVE_INTEGER_SCHEMA,
    NON_POSITIVE_INTEGER_SCHEMA,
    NULL_SCHEMA
)
from ._typing import (
    JSONSchema,
    JSONDefinitions,
    is_schema
)



import json
from abc import abstractmethod
from typing import IO, Optional

from wai.common.decorator import ensure_error_type

from ..error import JSONSerialisationError
from ..raw import RawJSONElement
from ..validator import JSONValidator


class JSONSerialisable:
    """
    Interface class for objects which can serialise themselves to JSON.
    """
    @abstractmethod
    def _serialise_to_raw_json(self) -> RawJSONElement:
        """
        Performs the actual serialisation.

        :return:    The JSON representation.
        """
        pass

    @ensure_error_type(JSONSerialisationError, "Error converting {self.__class__.__qualname__} to raw JSON: {0}")
    def to_raw_json(self, validate: bool = True) -> RawJSONElement:
        """
        Converts the state of this object to a raw JSON element.

        :param validate:    Whether to validate the serialised JSON if possible.
        :return:            The raw JSON.
        """
        # Get the raw JSON representation
        raw_json: RawJSONElement = self._serialise_to_raw_json()

        # Validate the JSON if we are capable
        if validate and isinstance(self, JSONValidator):
            self.validate_raw_json(raw_json)

        return raw_json

    @ensure_error_type(JSONSerialisationError, "Error converting {self.__class__.__qualname__} to JSON string: {0}")
    def to_json_string(self, indent: Optional[int] = None, validate: bool = True) -> str:
        """
        Serialises this object to a JSON-format string.

        :param indent:      The indent level to use for pretty-printing, or
                            None for compact representation.
        :param validate:    Whether to validate the serialised JSON if possible.
        :return:            The JSON string.
        """
        if indent is not None and indent < 0:
            raise JSONSerialisationError(f"Indent level for pretty-printing must be non-negative, "
                                         f"got {indent}")

        return json.dumps(self.to_raw_json(validate),
                          indent=indent,
                          separators=(',', ':') if indent is None else None)

    @ensure_error_type(JSONSerialisationError, "Error writing {self.__class__.__qualname__} to stream: {0}")
    def write_json_to_stream(self, stream: IO[str], indent: Optional[int] = None, validate: bool = True) -> None:
        """
        Writes this object as JSON to a string-stream.

        :param stream:      The stream to write to.
        :param indent:      The indent level to use for pretty-printing, or
                            None for compact representation.
        :param validate:    Whether to validate the serialised JSON if possible.
        """
        if indent is not None and indent < 0:
            raise JSONSerialisationError(f"Indent level for pretty-printing must be non-negative, "
                                         f"got {indent}")

        json.dump(self.to_raw_json(validate),
                  stream,
                  indent=indent,
                  separators=(',', ':') if indent is None else None)

    @ensure_error_type(JSONSerialisationError, "Error saving {self.__class__.__qualname__} to '{filename}': {0}")
    def save_json_to_file(self, filename: str, indent: Optional[int] = None, validate: bool = True) -> None:
        """
        Saves this object to the given file.

        :param filename:    The name of the file to save to.
        :param indent:      The indent level to use for pretty-printing, or
                            None for compact representation.
        :param validate:    Whether to validate the serialised JSON if possible.
        """
        with open(filename, 'w') as file:
            self.write_json_to_stream(file, indent, validate)



from ..schema import JSONSchema
from ._JSONValidator import JSONValidator


class BasicSchemaValidator(JSONValidator):
    """
    Basic wrapper for a JSON schema which presents it as a validator.
    """
    def __init__(self, schema: JSONSchema):
        self._schema = schema

    def _get_json_validation_schema(self) -> JSONSchema:
        return self._schema



from abc import ABC

from ..validator import JSONValidator
from ._JSONBiserialisable import JSONBiserialisable, SelfType


class JSONValidatedBiserialisable(JSONValidator, JSONBiserialisable[SelfType], ABC):
    """
    Utility interface that specifies that a class is both biserialisable
    with JSON and that that JSON is validated.
    """
    pass



from ._JSONPropertyError import JSONPropertyError


class OptionalDisallowed(JSONPropertyError):
    """
    Error type for when trying to set a property as optional,
    but it is not allowed.
    """
    pass



from typing import List, Iterable, Optional

from ...error import OfPropertySelectionError
from ...schema import any_of
from .._typing import PropertyValueType, Absent, OptionallyPresent
from ._OfProperty import OfProperty
from ._Property import Property


class AnyOfProperty(OfProperty):
    """
    Property which validates JSON that matches at least one of
    a number of sub-properties.
    """
    def __init__(self,
                 name: Optional[str] = None,
                 sub_properties: Iterable[Property] = tuple(),
                 *,
                 optional: bool = False,
                 default: OptionallyPresent[PropertyValueType] = Absent):
        super().__init__(
            name,
            sub_properties,
            schema_function=any_of,
            optional=optional,
            default=default
        )

    @classmethod
    def _choose_subproperty(cls, successes: List[bool]) -> int:
        # Search the list
        for i, success in enumerate(successes):
            # Return the first valid index found
            if success:
                return i

        raise OfPropertySelectionError(f"Value didn't match any sub-properties")



from typing import TypeVar

from ._JSONObject import JSONObject

SelfType = TypeVar("SelfType", bound='StrictJSONObject')


class StrictJSONObject(JSONObject[SelfType]):
    """
    Base class for JSON objects which don't support additional
    properties by default.
    """
    @classmethod
    def _additional_properties_validation(cls):
        return None



from ._JSONPropertyError import JSONPropertyError


class OfPropertySelectionError(JSONPropertyError):
    """
    Error for when an of-property (any-of, one-of, all-of) can't
    select a sub-property to match a value.
    """
    pass



from typing import Type, Optional

from .._typing import PropertyValueType, Absent, OptionallyPresent
from ._ProxyProperty import ProxyProperty


class JSONObjectProperty(ProxyProperty):
    """
    Property which validates a sub-configuration (essentially an object).
    """
    def __init__(self,
                 name: Optional[str] = None,
                 object_type: Optional[Type['JSONObject']] = None,
                 *,
                 optional: bool = False,
                 default: OptionallyPresent[PropertyValueType] = Absent):
        # Have to local-import JSONObject to avoid circular reference
        from .._JSONObject import JSONObject

        # If no object type specified, allow any object
        if object_type is None:
            object_type = JSONObject

        super().__init__(
            name,
            proxy=object_type,
            optional=optional,
            default=default
        )

    @property
    def object_type(self) -> Type['JSONObject']:
        """
        Gets the type of JSON object this property takes.
        """
        return self.proxy_type



from typing import Optional, Any

from ...schema import JSONSchema, TRIVIALLY_FAIL_SCHEMA
from .._typing import PropertyValueType, Absent, OptionallyPresent
from ._Property import Property


class RawProperty(Property):
    """
    Property which takes raw JSON values. Uses a schema for validation.
    """
    def __init__(self,
                 name: Optional[str] = None,
                 *,
                 schema: JSONSchema = TRIVIALLY_FAIL_SCHEMA,
                 optional: bool = False,
                 default: OptionallyPresent[PropertyValueType] = Absent):
        self._schema = schema

        super().__init__(name, optional=optional, default=default)

    def _get_json_validation_schema(self) -> JSONSchema:
        return self._schema

    def _validate_value(self, value: Any) -> PropertyValueType:
        # Perform schema validation
        self.validate_raw_json(value)

        return value



from abc import ABC, abstractmethod
from sys import maxsize
from typing import Iterable, Optional, List, Callable, Any, Iterator, Type

from ....error import JSONError, OptionalDisallowed
from ....raw import RawJSONElement
from ....serialise import JSONValidatedBiserialisable
from ....schema import JSONSchema, regular_array
from ....validator import StaticJSONValidator
from ..._typing import PropertyValueType
from .._Property import Property


class ArrayProxy(StaticJSONValidator, JSONValidatedBiserialisable['ArrayProxy'], ABC):
    """
    Class which acts like an array, but validates its elements using a property.
    """
    def __init__(self, initial_values: Optional[Iterable] = None):
        # The list values
        self._values: List[PropertyValueType] = []

        # Add any initial values
        if initial_values is not None:
            self.__iadd__(initial_values)

        # Make sure the list is long enough
        if len(self._values) < self.min_elements():
            raise JSONError(f"Array proxy initialised with {len(self._values)} elements "
                            f"when the minimum is {self.min_elements()}")

    @staticmethod
    @abstractmethod
    def element_property() -> Property:
        """
        Gets the property which is used to validate elements of the array.
        """
        pass

    @staticmethod
    @abstractmethod
    def min_elements() -> int:
        """
        The minimum-allowed length of this array.
        """
        pass

    @staticmethod
    @abstractmethod
    def max_elements() -> Optional[int]:
        """
        The maximum-allowed length of this array.
        """
        pass

    @staticmethod
    @abstractmethod
    def unique_elements() -> bool:
        """
        Whether the elements of this array have to be distinct
        from one another.
        """
        pass

    @staticmethod
    def specify(element_property: Property,
                min_elements: int = 0,
                max_elements: Optional[int] = None,
                unique_elements: bool = False) -> Type['ArrayProxy']:
        """
        Creates a specific array proxy-type.

        :param element_property:    The property to use to validate array elements.
        :param min_elements:        The minimum number of elements allowed in the array.
        :param max_elements:        The maximum number of elements allowed in the array.
        :param unique_elements:     Whether array elements have to be unique.
        :return:                    The array proxy-type matching the specification.
        """
        # Create the proxy-type as a closure
        class ClosureArrayProxy(ArrayProxy):
            @staticmethod
            def element_property() -> Property:
                return element_property

            @staticmethod
            def min_elements() -> int:
                return min_elements

            @staticmethod
            def max_elements() -> Optional[int]:
                return max_elements

            @staticmethod
            def unique_elements() -> bool:
                return unique_elements

        return ClosureArrayProxy

    def _serialise_to_raw_json(self) -> RawJSONElement:
        return [value.to_raw_json() if isinstance(value, JSONValidatedBiserialisable) else value
                for value in self._values]

    @classmethod
    def _deserialise_from_raw_json(cls, raw_json: RawJSONElement) -> 'ArrayProxy':
        return cls(raw_json)

    @classmethod
    def _get_json_validation_schema(cls) -> JSONSchema:
        return regular_array(
            cls.element_property().get_json_validation_schema(),
            cls.min_elements(),
            cls.max_elements(),
            cls.unique_elements()
        )

    def __init_subclass__(cls, **kwargs):
        # Make sure the sub-property isn't optional
        if cls.element_property().is_optional:
            raise OptionalDisallowed("Can't use optional sub-property with array proxies")

        # Make sure min_elements is not negative
        if cls.min_elements() < 0:
            raise JSONError(f"min_elements must be non-negative, got {cls.min_elements()}")

        super().__init_subclass__(**kwargs)

    # ------------ #
    # LIST METHODS #
    # ------------ #

    def append(self, value):
        # Make sure we're not already at max length
        if len(self._values) == self.max_elements():
            raise JSONError(f"Tried to append to list already of maximum size ({self.max_elements()})")

        # Make sure the unique-elements constraint isn't violated
        if self.unique_elements() and value in self:
            raise JSONError(f"Attempted to add non-unique element")

        self._values.append(self.element_property().validate_value(value))

    def clear(self):
        # Make sure clearing the list wouldn't violate the minimum size
        if self.min_elements() > 0:
            raise JSONError(f"Cannot clear array when minimum size "
                            f"({self.min_elements()}) is greater than zero")

        # Clear the key-list
        self._values.clear()

    def copy(self):
        return type(self)(self)

    def count(self, value) -> int:
        try:
            value = self.element_property().validate_value(value)
            return self._values.count(value)
        except Exception:
            return 0

    def extend(self, iterable: Iterable):
        for value in iterable:
            self.append(value)

    def index(self, value, start: int = 0, stop: int = maxsize) -> int:
        value = self.element_property().validate_value(value)
        return self._values.index(value, start, stop)

    def insert(self, index: int, value):
        # Make sure we're not already at max length
        if len(self._values) == self.max_elements():
            raise JSONError(f"Tried to insert into list already of maximum size ({self.max_elements()})")

        # Make sure the unique-elements constraint isn't violated
        if self.unique_elements() and value in self:
            raise JSONError(f"Attempted to insert non-unique element")

        self._values.insert(index, self.element_property().validate_value(value))

    def pop(self, index: int = -1):
        # Make sure we're not already at min length
        if len(self._values) == self.min_elements():
            raise JSONError(f"Tried to pop from list already of minimum size ({self.min_elements()})")

        return self._values.pop(index)

    def remove(self, value):
        self.pop(self.index(value))

    def reverse(self):
        self._values.reverse()

    def sort(self, *,
             key: Optional[Callable[[Any], Any]] = None,
             reverse: bool = False):
        self._values.sort(key=key, reverse=reverse)

    def __add__(self, x: List) -> List:
        return self[:] + x

    def __contains__(self, value) -> bool:
        return self.element_property().validate_value(value) in self._values

    def __delitem__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(ArrayProxy.__delitem__.__qualname__)

    def __eq__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(ArrayProxy.__eq__.__qualname__)

    def __getitem__(self, y):
        return self._values[y]

    def __ge__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(ArrayProxy.__ge__.__qualname__)

    def __gt__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(ArrayProxy.__gt__.__qualname__)

    def __iadd__(self, x: Iterable) -> 'ArrayProxy':
        for value in x:
            self.append(value)
        return self

    def __imul__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(ArrayProxy.__imul__.__qualname__)

    def __iter__(self) -> Iterator:
        return iter(self._values)

    def __len__(self) -> int:
        return len(self._values)

    def __le__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(ArrayProxy.__le__.__qualname__)

    def __lt__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(ArrayProxy.__lt__.__qualname__)

    def __mul__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(ArrayProxy.__mul__.__qualname__)

    def __ne__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(ArrayProxy.__ne__.__qualname__)

    def __repr__(self):
        # TODO: Implement
        raise NotImplementedError(ArrayProxy.__repr__.__qualname__)

    def __reversed__(self):
        return reversed(self._values)

    def __rmul__(self, *args, **kwargs):
        # TODO: Implement
        raise NotImplementedError(ArrayProxy.__rmul__.__qualname__)

    def __setitem__(self, index: int, value):
        # Validate the value
        value = self.element_property().validate_value(value)

        # Make sure the unique-elements constraint isn't violated
        if self.unique_elements() and value in self._values and self._values[index] != value:
            raise JSONError(f"Attempted to set element to non-unique element")

        self._values[index] = value

    def __str__(self):
        return str(self._values)
